/*
 * The MIT License
 *
 * Copyright 2012 Xavi Miró.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package net.sf.jinterflex;

import java.lang.annotation.Annotation;
import java.lang.reflect.*;
import java.util.Arrays;

/**
 * The dynamic proxy handler for the class calls of JInterflex.
 *
 * @author Xavi Miró
 */
final class ClassProxyHandler<T> implements InvocationHandler, IProxified {

    private final Class<?> realClass_;
    private final Class<T> classInterface_;
    private final String ctorName_;

    /**
     * Ctor.
     * @param iface the auxiliary interface which contains the method and field
     * accesses to do to the real class
     */
    public ClassProxyHandler(Class<T> iface)
    throws ClassNotFoundException {

        classInterface_ = iface;
        realClass_ = getRealClass(iface);
        ctorName_ = getCtorName(iface);
    }

    /**
     *
     */
    @Override
    public Object invoke(Object proxy, Method method, Object[] args)
    throws Throwable {

        String methodName = method.getName();
        if (methodName.equals(ctorName_)){
            Class<?> expectedRetType = classInterface_.getEnclosingClass();
            if (!method.getReturnType().equals(expectedRetType)){
                throw new IllegalArgumentException("The method " + methodName + " should return a " + expectedRetType + ", not a " + method.getReturnType());
            }
        }
        return invokeMethod(method, null, realClass_, args, ctorName_);
    }

    /**
     * Utility method useful for class handler or instance handler calls.
     */
    static <T> T invokeMethod(Method simulMethod, T obj, Class<?> clz, Object[] rawArgs, String ctorName)
    throws Throwable {

        T res = null;
        try {
            String methodName = simulMethod.getName();
            Class<?>[] rawParameterTypes = simulMethod.getParameterTypes();
            Class <?>[] parameterTypes = rawParameterTypes;
            Object[] args = rawArgs;
            if (rawArgs != null && rawParameterTypes != null){
                parameterTypes = Arrays.copyOf(rawParameterTypes, rawParameterTypes.length);
                args = Arrays.copyOf(rawArgs, rawArgs.length);
                modifyArgsAndTypes(args, parameterTypes);
            }
            if (ctorName != null && methodName.equals(ctorName)){
                Constructor ctor = clz.getConstructor(parameterTypes);
                res = (T)ctor.newInstance(args);
            }
            else if (simulMethod.getAnnotation(FieldAccess.class) != null){
                Field field = clz.getField(methodName);
                res = (T)field.get(null);
            }
            else {
                Method realMethod = clz.getMethod(methodName, parameterTypes);
                res = (T)realMethod.invoke(obj, args);
            }
            Class<?> retType = simulMethod.getReturnType();
            if (retType.isAnnotationPresent(JInterflexProxy.class)){
                res = (T)Proxy.newProxyInstance(ClassProxyHandler.class.getClassLoader(), new Class[]{retType},
                                                new InstanceProxyHandler(res, clz));
            }
        }
        catch (ClassNotFoundException ex) {
            throw new JInterflexException(getGeneralErrMsg(simulMethod, clz), ex);
        }
        catch (NoSuchMethodException ex) {
            throw new JInterflexException(getGeneralErrMsg(simulMethod, clz), ex);
        }
        catch (NoSuchFieldException ex) {
            throw new JInterflexException(getGeneralErrMsg(simulMethod, clz), ex);
        }
        catch (IllegalAccessException ex) {
            throw new JInterflexException(getGeneralErrMsg(simulMethod, clz), ex);
        }
        catch (InstantiationException ex) {
            throw new JInterflexException(getGeneralErrMsg(simulMethod, clz), ex);
        }
        catch (InvocationTargetException e){
            throw e.getCause();
        }

        return res;
    }

    /**
     *
     */
    private static String getGeneralErrMsg(Method simulMethod, Class<?> realClass){

       return "Error making call " + simulMethod.getName() + " on class " + realClass.getName();
    }

    /**
     *
     */
    private String getCtorName(Class<T> iface) {

        String res = null;

        JInterflexProxy ctorAt = iface.getEnclosingClass().getAnnotation(JInterflexProxy.class);
        res = ctorAt.ctorMethodName();

        return res;
    }

    /**
     *
     */
    private static void modifyArgsAndTypes(Object[] args, Class<?>[] parameterTypes)
    throws ClassNotFoundException {

        for (int j = 0; j < parameterTypes.length; ++j){
            Class<?> type = parameterTypes[j];
            if (type.isAnnotationPresent(JInterflexProxy.class)){
                for (Annotation annot: type.getDeclaredAnnotations()){
                    if (annot.annotationType().equals(JInterflexProxy.class)){
                        parameterTypes[j] = Class.forName(( (JInterflexProxy)annot ).realClassName());
                    }
                    IInstanceProxy proxyInstance = (IInstanceProxy)Proxy.getInvocationHandler(args[j]);
                    args[j] = proxyInstance.getRealObject();
                }
            }
        }
    }

    /**
     *
     */
    private Class<?> getRealClass(Class<T> iface) throws ClassNotFoundException {

        if (!iface.isInterface()){
            throw new IllegalArgumentException(iface.getSimpleName() + " is not an interface");
        }
        Class<?> encInterface = iface.getEnclosingClass();
        if (encInterface == null || !encInterface.isInterface()){
            throw new IllegalArgumentException("The interface " + iface + " has no enclosing interface");
        }
        if (!encInterface.isAnnotationPresent(JInterflexProxy.class)){
            throw new IllegalArgumentException("The enclosing interface of " + iface.getSimpleName() + " has no " + JInterflexProxy.class.getSimpleName() + " annotation defined");
        }
        JInterflexProxy proxyAnnot = encInterface.getAnnotation(JInterflexProxy.class);
        String realClassName = proxyAnnot.realClassName();

        return Class.forName(realClassName);
    }
}
