/*
 * The MIT License
 *
 * Copyright 2012 Xavi Miró.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.sf.jinterflex;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;

/**
 * Creates and manages proxy objects that can
 * read public fields and execute methods through reflection
 * using interfaces and annotations.
 *
 * @author Xavi Miró
 */
public final class JInterflex {

    /**
     * Private ctor to prevent building instances of it.
     */
    private JInterflex(){}

    /**
     * Creates a proxy class that implements a given auxiliary interface
     * in order to invoke methods and read public fields on another class
     * through reflection.
     * @param <T> the type of the interface <code>iface</code>.
     * @param iface the auxiliary interface that contains the methods and fields
     * that mimic the signatures of the class to be accessed.
     * @return the proxy class.
     * @throws IllegalArgumentException if the <code>iface</code> argument
     * does not follow the mandatory conditions to be used as a <code>JInterflex</code> proxy.
     */
    public static <T> T proxy(Class<T> iface) {

        T res = null;

        if (!iface.isInterface()){
            throw new IllegalArgumentException(iface.getName() + " must be an interface");
        }
        Class enclosingIface = iface.getEnclosingClass();
        if (enclosingIface == null){
            throw new IllegalArgumentException(iface.getName() + " must be an inner interface");
        }
        if (!enclosingIface.isAnnotationPresent(JInterflexProxy.class)){
            throw new IllegalArgumentException(iface.getName() +
                                               " must have an enclosing interface with a @" +
                                               JInterflexProxy.class.getSimpleName() +
                                               " annotation");
        }
        ClassLoader clzLoader = ClassProxyHandler.class.getClassLoader();
        try {
            res = (T)Proxy.newProxyInstance(clzLoader, new Class[]{iface}, new ClassProxyHandler<T>(iface));
        }
	catch (Exception ex) {
            res = (T)Proxy.newProxyInstance(clzLoader, new Class[]{iface}, new UnavailableProxyHandler(new JInterflexException("Error creating proxy of class " + iface.getName(), ex)));
        }

        return res;
    }

    /**
     * It returns true if the object <code>proxy</code> is available to be used, i.e. there were no errors during its creation. If this
     * methods returns <code>false</code>, there was an error so, trying to call
     * the methods of this object will cause <code>JInterflexExceptions</code> to be thrown<p/>
     * The method JInterflex#getException will return the exception that was
     * thrown during the creation of the proxy object.
     * @param proxy the object to check whether it is in valid state or not.
     * @return <code>true</code> if it's available.
     * @throws NullPointerException if the object <code>proxy</code> is <code>null</code>.
     * @throws IllegalArgumentException if the object <code>proxy</code> has not been
     * returned by the {@link JInterflex#proxy(java.lang.Class)} method.
     */
    public static boolean isProxyAvailable(Object proxy){

        boolean res = false;

        if (proxy == null){
            throw new NullPointerException("The argument cannot be null");
        }
        if (Proxy.isProxyClass(proxy.getClass())){
            InvocationHandler invHandler = Proxy.getInvocationHandler(proxy);
            if (!(invHandler instanceof IProxified)){
                throw new IllegalArgumentException("This object has not been created by " +
                                                   JInterflex.class.getSimpleName() +
                                                   ". Its invocation handler is a " + invHandler);
            }
            if (invHandler instanceof ClassProxyHandler){
                res = true;
            }
        }
        else {
            throw new IllegalArgumentException("This object has not been created by " +
                                               JInterflex.class.getSimpleName() +
                                               ". It is a " + proxy.getClass().getName());
        }
        return res;
    }

    /**
     * It returns the exception associated to the error that prevented the
     * creation of the proxy object.
     * @param proxy the proxy.
     * @return the exception of the error if there has been an error,
     * <code>null</code> otherwise.
     * @throws NullPointerException if the object <code>proxy</code> is <code>null</code>.
     * @throws IllegalArgumentException if the object <code>proxy</code> has not been
     * returned by the {@link JInterflex#proxy(java.lang.Class)} method.
     */
    public static Exception getProxyException(Object proxy){

        Exception res = null;
        InvocationHandler invHandler = Proxy.getInvocationHandler(proxy);
        if (( invHandler instanceof UnavailableProxyHandler)){
            UnavailableProxyHandler errProxy = (UnavailableProxyHandler)invHandler;
            res = errProxy.getException();
        }

        return res;
    }
}

